"""Module contains common constants."""
INQUIRERPY_KEYBOARD_INTERRUPT: str = "INQUIRERPY_KEYBOARD_INTERRUPT"

INQUIRERPY_POINTER_SEQUENCE: str = "\u276f"
INQUIRERPY_FILL_CIRCLE_SEQUENCE: str = "\u25c9"
INQUIRERPY_EMPTY_CIRCLE_SEQUENCE: str = "\u25cb"
INQUIRERPY_QMARK_SEQUENCE: str = "\u003f"
